import json

import iota_client


def check_installation():
    '''Getting Started
    https://client-lib.docs.iota.org/docs/libraries/python/getting_started
    '''
    print('\n[ Installation ]')
    print(iota_client.__doc__)
    print(dir(iota_client))


def check_network():
    print('\n[ Network ]')
    # client will connect to testnet by default
    client = iota_client.Client()
    print(json.dumps(client.get_info(), indent=4))


def main():
    check_installation()
    check_network()


if __name__ == '__main__':
    main()



from base64 import b64decode, b64encode
from datetime import datetime
import json

from Crypto import Random
from Crypto.Cipher import PKCS1_v1_5
from Crypto.Hash import SHA
from Crypto.PublicKey import RSA
import magic


def cid_decrypt(encrypted_cid, private_key):
    key = RSA.importKey(b64decode(private_key))
    dsize = SHA.digest_size
    sentinel = Random.new().read(15 + dsize)
    cipher = PKCS1_v1_5.new(key)
    message = cipher.decrypt(b64decode(encrypted_cid), sentinel)
    return message


def cid_encrypt(cid, public_key):
    pubkey = RSA.importKey(b64decode(public_key))
    cipher = PKCS1_v1_5.new(pubkey)

    # without padding digest
    message = b64encode(cipher.encrypt(cid.encode('utf-8')))

    # with padding digest
    # h = SHA.new(cid.encode('utf-8'))
    # print(h.digest())
    # message = b64encode(cipher.encrypt((cid).encode('utf-8') + h.digest()))

    return message


# Utils

def create_lite_metadata_from_cid(cid, ipfs):
    content_buffer = ipfs.cat(cid)

    mime_type = magic.from_buffer(content_buffer, mime=True)
    size_in_bytes = len(content_buffer)
    file_name = '{name}.{ext}'.format(
        name=int(datetime.now().timestamp()),
        ext=mime_type.split('/')[1]
    )

    return [{
        'cid': cid,
        'filename': file_name,
        'mimeType': mime_type,
        'size': size_in_bytes
    }]


def create_lite_metadata_from_file(filepath, ipfs):
    cid = ipfs.add(filepath)['Hash']
    return create_lite_metadata_from_cid(cid, ipfs)


def add_metadata_to_ipfs(
        ipfs_http_client,
        metadata,
        metadata_filepath='/tmp/metadata.json'):
    with open(metadata_filepath, 'w') as f:
        json.dump(metadata, f)
    cid = ipfs_http_client.add(metadata_filepath)['Hash']
    return cid


def create_transaction(
        niota,
        ipfs_http_client,
        cid,
        sender_peer_id,
        receiver_address,
        receiver_public_key,
        tag):
    metadata = create_lite_metadata_from_cid(cid, ipfs_http_client)
    meta_cid = add_metadata_to_ipfs(ipfs_http_client, metadata)
    message = {
        'pid': sender_peer_id,
        'cid': cid_encrypt(meta_cid, receiver_public_key).decode('utf-8')
    }
    return niota.propose_transaction(
        tag,
        message,
        value=0,
        address=receiver_address)



import os

from niota.chrysalis import IotaClient
from niota.signature import generate_key_pair


'''
Create a testing environment the same as Storage Backend.

Storage Backend does not use niota currently. The goal of this script is to
create a minimum environment for testing niota for Storage Backend
in the future:

    1. Ensure Storage Backend IOTA works here. (done)
    2. Ensure niota works here.
    3. Ensure niota to be compatible or with minimal effort for integration.

utils/{iota_chrysalis, signature}.py are copied from Storage Backend
without any change.

Usage
    $ python3 utils/hello_iota_storage_backend.py
'''

Mainnet_explorer = 'https://explorer.iota.org/mainnet'


def main():
    private_key, public_key = generate_key_pair()
    iota = IotaClient(private_key, public_key)

    raw_cid = 'cidhash'
    service_message = 'hello world'
    message_id, index = iota.create_message(
        raw_cid=raw_cid,
        provider='HelloProtocol',
        service_message=service_message,
    )

    print('Message ID:', message_id)
    print('Message URL:', os.path.join(Mainnet_explorer, 'message', message_id))
    print('Index:', index)

    message_data = iota.get_message_data(message_id)
    iota.verify_message_data(message_data)
    print('Message verified successfully')


if __name__ == '__main__':
    main()



'''
This example was from e02_send_hello.py of IOTA Python workshop.
It creates a simple transaction with a custom message/tag
There is no value attached to this transaction.
'''

import json
import logging

from iota import Address, Iota, ProposedTransaction, Tag, TryteString
from iota.adapter import HttpAdapter
from iota.crypto.types import Seed
from iota.crypto.addresses import AddressGenerator


logger = logging.getLogger(__name__)


class NumbersIOTA():
    def __init__(self):
        self.seed = None
        self.address = None
        self.api = Iota(
            HttpAdapter('https://nodes.thetangle.org:443', timeout=20)
        )
        self.base_tangle_add = 'https://utils.iota.org/transaction/{}'

    def generate_seed(self):
        # Note that we don't need a seed to send 0 value transactions since
        # these transactions are not signed, we can publish to any address
        self.seed = Seed.random()

    def generate_address(self, security_level=2):
        # security level is defined during generator init
        if self.seed is None:
            self.generate_seed()

        try:
            generator = AddressGenerator(
                seed=self.seed, security_level=security_level
            )
            self.address = generator.get_addresses(0, 1)[0]  # index, count
            logger.info("New address generated.")
        except Exception as e:
            logger.error("Fail to generate new address")
            print(e)
            raise

    def get_transaction(self, trans_hash):
        """ Allow users to input a trans_hash hash and get one transaction"""

        return self.get_transactions(trans_hash)

    def get_transactions_by_add(self, _addresses):
        """ Allow users to input a address or a list of addresses """

        if isinstance(_addresses, str):
            addresses = [_addresses]
        else:
            addresses = _addresses
        response = self.api.find_transaction_objects(
            addresses=addresses)
        if len(response["transactions"]) == 1:
            return response["transactions"][0]
        else:
            return response["transactions"]

    def get_tranaction_by_tag(self, _tag):
        """ Allow users to input tag or a list of tags """
        if isinstance(_tag, str):
            tag = [_tag]
        else:
            tag = _tag
        response = self.api.find_transaction_objects(tags=tag)
        if len(response["transactions"]) == 1:
            return response["transactions"][0]
        else:
            return response["transactions"]

    def get_transactions(self, _trans_hashes):
        """ Allow users to input a trans_hash or a list of hashes """

        if isinstance(_trans_hashes, str):
            trans_hashes = [_trans_hashes]
        else:
            trans_hashes = _trans_hashes
        response = self.api.get_transaction_objects(trans_hashes)
        if len(response["transactions"]) == 1:
            return response["transactions"][0]
        else:
            return response["transactions"]

    def get_message(self, transaction, conv_to_dict=True):
        """ Retuen decoded message of a transaction """
        _message = transaction.signature_message_fragment.decode(
            errors='ignore')
        if conv_to_dict:
            try:
                return json.loads(_message)
            except Exception:
                self.logger.error('Fail to convert message to dict')
        return _message

    def propose_transaction(
            self,
            tag,
            _message,
            value=0,
            address=None,
            provider='NUMBERSPROTOCOL'):
        """ Send data to IOTA."""

        # If there is no specified address, it will use self.address
        # or generate a new one if self.address is also None
        if address is None:
            address = self.address
            if self.address is None:
                self.generate_address()
        else:
            self.address = address
        try:
            if isinstance(_message, dict):
                _message['provider'] = provider
                message = _message
            else:
                message = {
                    'provider': provider,
                    'content': _message
                }
            message = json.dumps(message)
        except Exception as e:
            logger.error("Fail to convert dict to json.")
            print(e)
            raise
        try:
            tx = ProposedTransaction(
                address=Address(self.address),
                message=TryteString.from_unicode(message),
                value=0,
                tag=Tag(tag),
            )
            response = self.api.send_transfer([tx])
            return self.base_tangle_add.format(response['bundle'][0].hash)

        except Exception as e:
            logger.error("Fail to propose and run transaction")
            print(e)
            raise



from datetime import datetime
import hashlib
import json
import logging
import sys

import ipfshttpclient
from iota import TryteString

from .core import NumbersIOTA


logger = logging.getLogger(__name__)
niota = NumbersIOTA()


def parse_json(fname):
    with open(fname, "r", encoding='utf-8') as data_file:
        data = json.load(data_file)

    return data


# create cid
def create_cid(filepath, ipfs):
    cid = ipfs.add(filepath)['Hash']

    return cid


def create_tag(cid):
    cid_tryte = TryteString.from_unicode(cid)[0:24]
    tag = "NBS" + str(cid_tryte)

    return tag


def create_update_tag(input_tag):
    tag_length = len(input_tag)
    if tag_length < 27:
        extra = 27 - tag_length
        extra_value = extra * "9"
        tag = input_tag + extra_value
    else:
        tag = input_tag
    return tag


def create_ida_message(
        ida_cid='',
        ida_mid='',
        ida_sha256sum='',
        service_message=''):
    timestamp = int(datetime.now().timestamp())

    return {
        'timestamp': timestamp,
        'ida_cid': ida_cid,
        'ida_mid': ida_mid,
        'ida_sha256sum': ida_sha256sum,
        'service_message': service_message,
    }


def create_ida_transaction(
        ida_cid='',
        ida_mid='',
        ida_sha256sum='',
        provider='NUMBERSPROTOCOL',
        service_message=''):

    tag = create_tag(ida_cid)
    message = create_ida_message(
        ida_cid=ida_cid,
        ida_mid=ida_mid,
        ida_sha256sum=ida_sha256sum,
        service_message=service_message)
    value = 0
    full_add = niota.propose_transaction(
        tag, message, value=value, provider=provider)
    transaction_hash = full_add[-81:]
    logger.info("Transaction sent to the tangle!")
    logger.info(full_add)
    logger.info("Transaction Hash")
    logger.info(transaction_hash)
    logger.info("TAG")
    logger.info(tag)

    return {
        'tag': tag,
        'transaction': transaction_hash,
        'cid': ida_cid,
    }


def create_record_transaction(
        tag='',
        ida_cid='',
        ida_mid='',
        ida_sha256sum='',
        provider='NUMBERSPROTOCOL',
        service_message=''):

    message = create_ida_message(
        ida_cid=ida_cid,
        ida_mid=ida_mid,
        ida_sha256sum=ida_sha256sum,
        service_message=service_message)
    value = 0
    full_add = niota.propose_transaction(
        tag, message, value=value, provider=provider)
    transaction_hash = full_add[-81:]
    logger.info("Transaction sent to the tangle!")
    logger.info(full_add)
    logger.info("Transaction Hash")
    logger.info(transaction_hash)
    logger.info("TAG")
    logger.info(tag)

    return {
        'tag': tag,
        'transaction': transaction_hash,
        'cid': ida_cid,
    }


def create_transaction_from_json(json_file):
    _json_data = parse_json(json_file)
    json_data = json.dumps(_json_data, sort_keys=True).encode('utf-8')
    json_hash = hashlib.sha256(json_data).hexdigest()

    ipfs = ipfshttpclient.connect('/ip4/127.0.0.1/tcp/5001')
    cid = create_cid(json_file, ipfs)

    return create_ida_transaction(
        ida_cid=cid,
        ida_sha256sum=json_hash,
        provider='NUMBERSIDA',
        service_message='creation transaction')


def create_update_transaction_from_json(json_file, input_tag):
    _json_data = parse_json(json_file)
    json_data = json.dumps(_json_data, sort_keys=True).encode('utf-8')
    json_hash = hashlib.sha256(json_data).hexdigest()

    ipfs = ipfshttpclient.connect('/ip4/127.0.0.1/tcp/5001')
    cid = create_cid(json_file, ipfs)

    return create_record_transaction(
        tag=input_tag,
        ida_cid=cid,
        ida_sha256sum=json_hash,
        provider='NUMBERSIDA',
        service_message='update transaction')


def search(input_tag):
    hash_list = []
    transaction = niota.get_tranaction_by_tag(input_tag)
    # if only one transaction hash
    # print(transaction.hash)

    # if more than one transaction hash per tag
    for i in transaction:
        print(i.hash)
        hash_list.append(i.hash)

    # return transaction.hash
    return hash_list


def main():
    action = sys.argv[1]
    assert action in ['--first', '--update', '--search'], \
        'Action is not one of --first, --udpdate or --search: ' + action

    if action == '--first':
        filename = sys.argv[2]
        create_transaction_from_json(filename)

    elif action == '--update':
        filename = sys.argv[2]
        input_tag = sys.argv[3]

        tag = create_update_tag(input_tag)
        create_update_transaction_from_json(filename, tag)

    elif action == '--search':
        tag = sys.argv[2]

        search(tag)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import argparse
import json
import logging

import ipfshttpclient

from niota import ipfs
from niota import NumbersIOTA


logger = logging.getLogger(__name__)
niota = NumbersIOTA()


def load_config(config_filepath='iota_config.json'):
    with open(config_filepath) as f:
        return json.load(f)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--cid', '-c', required=True, help='Image CID')
    args = parser.parse_args()

    logger.info('Note: This tool requires a local IPFS daemon.')
    logger.info('Image CID: ' + args.cid)

    # Init
    niota_instance = niota.NumbersIOTA()
    ipfs_http_client = ipfshttpclient.connect('/ip4/127.0.0.1/tcp/5001')

    # Transaction
    config = load_config()
    transaction_url = ipfs.create_transaction(
        niota_instance,
        ipfs_http_client,
        args.cid,
        config['sender_peer_id'],
        config['receiver_address'],
        config['receiver_public_key'],
        config['tag'],
    )
    logger.info('Transaction URL: {}'.format(transaction_url))


if __name__ == "__main__":
    main()



import hashlib
import json
import logging

from .core import NumbersIOTA


logger = logging.getLogger(__name__)


class NumbersVerify():
    def __init__(self, logger):
        self.niota = NumbersIOTA()

    def verify(
            self, trans_hash, verify_json, keys=['timestamp'],
            verify_hash=True, verify_provider=True, verify_sig=True):
        """
        Thie verify module will get the transaction message from IOTA and
        compare the following
            1. if the provider is NUMBERSPROTOCOL
            2. if the hash of the JSON data matches
            3. if the data specified by keys matches
            4. if the signature is valid

        Please SORT verify_json before sending it to the verify module.
        """

        transaction = self.niota.get_transaction(trans_hash)
        if isinstance(transaction, list) and len(transaction) == 0:
            logger.error('No transaction found')
            raise
        message = self.niota.get_message(transaction)

        if verify_provider:
            try:
                provider = message['provider']
            except KeyError:
                logger.warning('No provider found')
                return False
            if provider != 'NUMBERSPROTOCOL':
                logger.error('provider verification fails.')
                return False

        if verify_hash:
            try:
                registered_hash = message['data_hash']
            except KeyError:
                logger.warning('No data_hash found')
                return False
            logger.info('Registered hash = %s' % registered_hash)

            encoded_data = json.dumps(
                verify_json, sort_keys=True).encode('utf-8')
            verify_hash = hashlib.sha256(encoded_data).hexdigest()
            logger.info('Verify hash = %s' % verify_hash)

            if verify_hash != registered_hash:
                logger.error('data_hash verification fails.')
                return False

        if verify_sig:
            # TODO
            pass

        for verify_key in keys:
            try:
                ledger_data = message[verify_key]
            except KeyError:
                logger.warning(
                    'No %s found in ledger message' % verify_key)
                return False
            try:
                verify_data = verify_json[verify_key]
            except KeyError:
                logger.warning(
                    'No %s found in input JSON data' % verify_key)
                return False
            if ledger_data != verify_data:
                logger.error('%s verification fails.' % verify_key)
                return False
        logger.info('Numbers NIota Verification passed!')
        return True



import base64
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from typing import Union


def load_private_key(pem_private_key: str) -> rsa.RSAPrivateKeyWithSerialization:
    pem_private_key_bytes = pem_private_key.encode('utf-8')
    return serialization.load_pem_private_key(pem_private_key_bytes, password=None)


def load_public_key(pem_public_key: str) -> rsa.RSAPublicKeyWithSerialization:
    pem_public_key_bytes = pem_public_key.encode('utf-8')
    return serialization.load_pem_public_key(pem_public_key_bytes)


def export_private_key(private_key: rsa.RSAPrivateKeyWithSerialization) -> str:
    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return pem.decode('utf-8')


def export_public_key(private_key: rsa.RSAPrivateKeyWithSerialization) -> str:
    pem = private_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return pem.decode('utf-8')


def get_pem_public_key(pem_private_key):
    private_key = load_private_key(pem_private_key)
    public_key = private_key.public_key()
    pem_public_key = export_public_key(public_key)
    return pem_public_key


def generate_key_pair():
    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend()
    )
    return private_key, private_key.public_key()


def sign_message(
    private_key: Union[str, rsa.RSAPrivateKeyWithSerialization],
    message: str
):
    if isinstance(private_key, str):
        private_key = load_private_key(private_key)
    signature = private_key.sign(
        message.encode('utf-8'),
        padding.PSS(
            mgf=padding.MGF1(hashes.SHA256()),
            salt_length=padding.PSS.MAX_LENGTH
        ),
        hashes.SHA256()
    )
    base64_signature = base64.b64encode(signature).decode('utf-8')
    return base64_signature


def verify_signature(
    public_key: Union[str, rsa.RSAPublicKeyWithSerialization],
    base64_signature: str,
    message: str,
    raise_exception=True,
):
    if isinstance(public_key, str):
        public_key = load_public_key(public_key)
    signature = base64.b64decode(base64_signature.encode('utf-8'))
    try:
        public_key.verify(
            signature,
            message.encode('utf-8'),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
    except InvalidSignature:
        if raise_exception:
            raise
        return False
    else:
        return True



from datetime import datetime
import json
from typing import Dict

import iota_client

from niota import signature


class IotaClient():
    MAINNET_NODE_URLS = [['https://chrysalis-nodes.iota.org']]

    def __init__(self, private_key, public_key):
        self.private_key = private_key
        self.public_key = public_key
        self.client = iota_client.Client(
            nodes_name_password=self.MAINNET_NODE_URLS,
            node_sync_disabled=True
        )

    def create_message(
        self, raw_cid, ida_cid='', ida_mid='', ida_sha256sum='',
        metadata_cid='', provider='NUMBERSPROTOCOL', service_message=''
    ):
        index = self._create_index(raw_cid)
        ida_message = self._create_ida_message(
            raw_cid=raw_cid,
            ida_cid=ida_cid,
            ida_mid=ida_mid,
            ida_sha256sum=ida_sha256sum,
            metadata_cid=metadata_cid,
            provider=provider,
            service_message=service_message,
        )
        message_id = self._broadcast_message(index, ida_message)
        return message_id, index

    def get_message_data(self, message_id: str):
        message = self.client.get_message_data(message_id)
        data_int_array = message['payload']['indexation'][0]['data']
        data_str = bytearray(data_int_array).decode()
        return data_str

    def get_message_ids_from_index(self, index: str):
        response = self.client.find_messages([index])
        message_ids = [i['message_id'] for i in response]
        return message_ids

    def verify_message_data(self, data_str: str):
        data_with_signature = json.loads(data_str)
        data = data_with_signature['data']
        base64_signature = data_with_signature['signature']
        signature.verify_signature(self.public_key, base64_signature, data)

    def _broadcast_message(self, index: str, data: Dict) -> str:
        serialized_data = json.dumps(data, sort_keys=True)
        base64_signature = signature.sign_message(self.private_key, serialized_data)
        serialized_data_with_signature = json.dumps(
            {
                'data': serialized_data,
                'signature': base64_signature,
            }
        )
        message_id_indexation = self.client.message(
            index=index,
            data_str=serialized_data_with_signature
        )
        return message_id_indexation['message_id']

    def _create_ida_message(self, raw_cid, ida_cid, ida_mid, ida_sha256sum, metadata_cid, provider, service_message):
        timestamp = int(datetime.now().timestamp())
        ida_message = {
            'timestamp': timestamp,
            'raw_cid': raw_cid,
            'ida_cid': ida_cid,
            'ida_mid': ida_mid,
            'ida_sha256sum': ida_sha256sum,
            'metadata_cid': metadata_cid,
            'provider': provider,
            'service_message': service_message,
        }
        return ida_message

    def _create_index(self, ida_mid, prefix='Authmedia'):
        return f'{prefix}-{ida_mid}'



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="niota",
    version="0.6.1",
    author="Numbers Co., Inc",
    author_email="hi@numbersprotocol.io",
    description="Numbers PyOTA Wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/numbersprotocol/niota",
    packages=setuptools.find_packages(),
    license='GPLv3',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "cryptography>=3.4.6",
        "ipfshttpclient>=0.6.0",
        "PyOTA>=2.1.0",
        "pycryptodome>=3.10.1",
        "python-magic>=0.4.18",
    ],
    python_requires=">=3.6",
)
